# -*- coding: utf-8 -*-
""" Auxiliary functions to load and analyze EVA 2.0 data.
"""

import numpy
import csv
import matplotlib.pyplot as pyplot

def read_daResults(handle):
    """ Parse daResults.dat file and extract the raw data.
    """
    # skip the first two rows: the header and the empty string
    for i in range(2): handle.next()
    # create a list of lines, each line being a list itself
    lines = []
    # read the lines
    for line in handle:
        split_line = [float(x) for x in line.split()]
        lines.append(split_line)
    # make the lines list a numpy array
    lines_array = numpy.array(lines)
    return lines_array

def _get_nearest_number(array, value):
    """ Find the number in a ascending sorted array that is nearest to the
    given value and return it and its index.
    """
    idx = numpy.searchsorted(array, value)
    if abs(array[idx] - value) < abs(array[idx-1] - value):
        return array[idx], idx
    else:
        return array[idx-1], idx-1

class Injection():
    """This class stores the information about one injection.
    """
    def __init__(self):
        return
    
    def assign_data(self, data, baseline_time=20):
        """ Get the data, find the actual points of injections start/end
        based on approximate values from the injections.csv, slice the data.
        Also get the previous X datapoints for baseline calculation.
        """
        # find the neasrest points
        self.association_start, ass_start_index = _get_nearest_number(data[:,1], self.association_start)
        self.association_end, ass_end_index = _get_nearest_number(data[:,1], self.association_end)
        self.dissociation_end, dis_end_index = _get_nearest_number(data[:,1], self.dissociation_end)
        # assign the data
        self.data = data[ass_start_index:dis_end_index, :]
        # assign the baseline data
        self.baseline_data = data[ass_start_index - baseline_time: ass_start_index, :]
        # and assign the number of channels
        self.num_channels = (self.data.shape[1] - 4) / 4 # first 4 columns are composite
        self.channel_mask = [1]*self.num_channels # all channels are on by default
        # apply the channel_mask to create mean_signal and mean_baseline
        self.apply_channel_mask()
        # and create the timeline
        self.timeline = self.data[:,1]
        return
    
    def apply_channel_mask(self):
        """ Recalculate the mean signal intensity based on the channel mask.
        """
        self.mean_signal = numpy.zeros(self.data.shape[0])
        # add the channels up
        for channel, mask in enumerate(self.channel_mask):
            self.mean_signal = self.mean_signal + self.data[:, channel+4]*mask
        # and average them
        self.mean_signal = self.mean_signal / sum(self.channel_mask)
        # do the same with the baseline
        self.mean_baseline = numpy.zeros(self.baseline_data.shape[0])
        for channel, mask in enumerate(self.channel_mask):
            self.mean_baseline = self.mean_baseline + self.baseline_data[:, channel+4]*mask
        self.mean_baseline = self.mean_baseline / sum(self.channel_mask)
        return
        
    def plot(self):
        """ Plot the injection if the data is assigned.
        """
        # need to take into account the channel_mask
        pyplot.plot(self.timeline, self.mean_signal)
        pyplot.show()
        return
    
    def plot_channels(self):
        """ Plot the channels separately.
        """
        for channel in range(self.num_channels):
            if self.channel_mask[channel]:
                pyplot.plot(self.timeline, self.data[:, channel+4])
        pyplot.legend([str(channel+1) for channel in range(self.num_channels) if self.channel_mask[channel]], loc='lower right')
        pyplot.show()
        return
    
    def zero_align(self):
        """ Align the data to start from 0 time and 0 level.
        Also generate separate phases: association and dissociation.
        """
        # calculate the baseline level and the mean timestep in the baseline
        baseline = self.mean_baseline.mean()
        timestep = (self.baseline_data[1:, 1] - self.baseline_data[:-1,1]).mean()
        # create the timeline
        self.zero_timeline = numpy.zeros(self.mean_signal.shape[0] + 1)
        # first timepoint is 0, second is step, next - the values from the timeline array
        self.zero_timeline[1:] = self.timeline - self.timeline[0] + timestep
        # create the signal array
        self.zero_mean_signal = numpy.zeros(self.mean_signal.shape[0] + 1)
        # first value is 0, then go the actual values normalized to baseline
        self.zero_mean_signal[1:] = self.mean_signal - baseline
        # extract association
        association_length = numpy.where(self.timeline == self.association_end)[0][0]
        self.ass_timeline = self.zero_timeline[:association_length]
        self.association = self.zero_mean_signal[:association_length]
        # and dissociation phases
        self.dis_timeline = self.zero_timeline[association_length:] - self.zero_timeline[association_length]
        self.dissociation = self.zero_mean_signal[association_length:]
        return

def parse_injections_file(handle, data=None, baseline_time=20):
    """ Read-in the file with injections descriptions
    and return a list of Injection objects with assigned properties.
    """
    injections = []
    reader = csv.DictReader(handle, skipinitialspace=True)
    for line in reader:
        injection = Injection()

        injection.date = line["date"]
        injection.ligand = line["ligand"]
        injection.analyte = line["analyte"]
        injection.analyte_concentration = float(line["analyte_concentration"])
        injection.injection_number = int(line["injection_number"])
        injection.association_start = float(line["association_start"])
        injection.association_end = float(line["association_end"])
        injection.dissociation_end = float(line["dissociation_end"])
        
        if data is not None:
            injection.assign_data(data, baseline_time=baseline_time)
        
        injections.append(injection)
    return injections

